package othello.view;

import java.rmi.* ;
import java.rmi.server.UnicastRemoteObject;
import java.util.Observer ;
import java.util.Observable ;

import othello.model.RemoteGameModelInterface;
import othello.model.Pair;
import othello.model.GameConstants;
import othello.observation.* ;

/** The animator object caches the state of the board
 *  and animates transitions between states of the board.
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class Animator extends UnicastRemoteObject
implements RemoteConcurrentObserver, GameConstants
{
	private static final long serialVersionUID = 8361501772785269040L;

	// If Java supported multiple inheritance then it would
	// extend both java.util.Observable and java.rmi.server.UnicastRemoteObject.
	// Since that is not possible, we delegate to a private Observable object.
	private static class MyObservable extends Observable  {
		void setChangedBackDoor() { super.setChanged() ; }
	}
	
	private final MyObservable observableDelegate = new MyObservable() ;

	private int [][] square = new int[ROWS][COLS] ;
	{
		for( int r=0 ; r<ROWS ; ++r ) {
			for( int c=0 ; c<COLS ; ++c ) {
				square[r][c] = EMPTY ; } }
	}

	public final int QUIET_STATE = 0 ;
	public final int FLYING_STATE = 1 ;
	public final int FLIPPING_STATE = 2 ;

	private int state = QUIET_STATE ;
	private int flippingRow ;
	private int flippingCol ;
	private double amountOfFlip ;
	private double flyingRow ;
	private double flyingCol ;
	private int newColour ;
	private int gameState = RemoteGameModelInterface.UNINITIALIZED ;
	private int myColour = EMPTY ;

	private RemoteGameModelInterface gameModel = null ;

	public Animator( ) throws RemoteException {
		super() ;
		System.err.println("Animator adds itself "+this+" as observer");
	}

	public void setGameModel( RemoteGameModelInterface gameModel ) {
		synchronized( this) { this.gameModel = gameModel ; }

		try {
			gameModel.addObserver(this);
			for( int r = 0 ; r < ROWS ; ++r ) 
				for( int c = 0 ; c < COLS ; ++c )
					synchronized (this) {
						square[r][c] = gameModel.getPieceAt(r, c) ; }
			setState( gameModel.getGameState() ) ;
		} catch (RemoteException e) {
			e.printStackTrace();
			System.err.println( "Lost communication with server" );
		}
	}

	public void setColour( int colour ) {
		synchronized(this) { myColour = colour ; }
		setChanged() ; notifyObservers() ;
	}
	
	synchronized private int getColour() { return myColour ; }

	synchronized private void setState( int state ) {
		gameState = state ;
	}
	
	public void addObserver( Observer observer ) {
		observableDelegate.addObserver( observer ); }

	public void update( RemoteCountMonitorInterface countMonitor, Object arg ) {
		Pair pair = (Pair) arg ;
		int row = pair==null ? 0 : pair.getRow() ;
		int col = pair==null ? 0 : pair.getCol() ;
		int oldColour ; synchronized( this ) {
			oldColour = square[ row ][ col ] ; }
		int newColour  ;
		try {
			synchronized(this) { gameState = gameModel.getGameState() ; }
			newColour = gameModel.getPieceAt( row, col ) ; }
		catch( RemoteException e ) {
			System.err.println( "Lost communication with server" );
			e.printStackTrace( System.err );
			return ; }
		if( pair == null ) { // Only a change of state
			setChanged(); notifyObservers();
			try {
				countMonitor.decrement();
			} catch (RemoteException e) {
				e.printStackTrace();
				System.err.println( "Lost communication with server" );
			} }
		else if( oldColour == EMPTY ) {
			// This is a new piece being lain.
			animateFlyIn( row, col, newColour, countMonitor ) ;
		}
		else {
			// This is a flip.
			animateFlip( row, col, newColour, countMonitor ) ;
		}
	}

	public void move( final int row, final int col ) {
		if( gameModel == null ) return ;
		// Move off this thread in case this is the GUI thread.
		Thread thread = new Thread(){		
			@Override public void run() {
				try {
					// TODO Do something with the result.
					gameModel.move( row, col, getColour() ) ;
				} catch (RemoteException e) {
					System.err.println( "Lost communication with server" );
					e.printStackTrace(); } } };
		thread.start();	
	}

	synchronized public int getPieceAt( int r, int c ) { return square[r][c] ; }

	synchronized public int getState() { return state ; }

	/** Flipping amount varies from 0 (not at all) to 1 (fully flipped) */
	synchronized public double getFlippingAmount() { return amountOfFlip ; }

	synchronized public int getFlippingRow() { return flippingRow ; }

	synchronized public int getFlippingCol() { return flippingCol ; }

	synchronized public double getFlyingRow() { return flyingRow ; }

	synchronized public double getFlyingCol() { return flyingCol ; }

	synchronized public int getNewColour() { return newColour ; }

	private void animateFlyIn( final int row, final int col,
			final int newColour,
			final RemoteCountMonitorInterface counter ) {
		class FlyThread extends Thread {
			@Override public void run() {
				int iterations ;
				synchronized( Animator.this ) {
					state = FLYING_STATE ;
					Animator.this.newColour = newColour ;
					if( newColour == WHITE ) {
						flyingRow = -2 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+row) ; }
					else {
						flyingRow = ROWS + 1 ;
						flyingCol = COLS / 2.0 ;
						iterations = 10*(2+9-row) ; } }

				double dr = (row - flyingRow) / iterations ;
				double dc = (col - flyingCol) / iterations ;

				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						flyingRow += dr ;
						flyingCol += dc ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers();
				try {
					counter.decrement(); }
				catch( RemoteException e ) {
					e.printStackTrace( System.err );
					System.err.println( "Lost communication with server" ); } }
		}
		Thread t = new FlyThread() ;
		t.start();
	}

	private void animateFlip( final int row, final int col,
			final int newColour,
			final RemoteCountMonitorInterface counter ) {
		class FlipThread extends Thread {
			@Override public void run() {
				synchronized( Animator.this ) {
					state = FLIPPING_STATE ;
					Animator.this.newColour = newColour ;
					square[ row ][ col ] = EMPTY ;
					flippingRow = row ;
					flippingCol = col ;
					amountOfFlip = 0.0 ; }

				int iterations = 50 ;
				double delta =  1 / (double)iterations ;
				for( int i=0 ; i<iterations ; ++i ) {
					synchronized( Animator.this ) {
						amountOfFlip += delta ; 
						setChanged() ; }
					notifyObservers();
					try{ sleep(20) ; }
					catch( InterruptedException e ) {}
				}

				synchronized( Animator.this ) {
					state = QUIET_STATE ;
					square[ row ][ col ] = newColour ;
					setChanged() ; }
				notifyObservers();
				try {
					counter.decrement(); }
				catch( RemoteException e ) {
					e.printStackTrace( System.err );
					System.err.println( "Lost communication with server" ); } }
		}
		Thread t = new FlipThread() ;
		t.start();
	}

	private void setChanged() { observableDelegate.setChangedBackDoor() ; }

	private void notifyObservers() { observableDelegate.notifyObservers(); }

	synchronized public String getColourString() {
		if( myColour == BLACK ) return "You are black." ;
		if( myColour == WHITE ) return "You are white." ;
		return "" ;
	}

	synchronized public String stateMessage() {
		if( state != QUIET_STATE ) return "..." ;
		switch( gameState ) {
		case RemoteGameModelInterface.UNINITIALIZED:
			return "Waiting to begin." ;
		case RemoteGameModelInterface.TIE_GAME:
			return "Tied game." ;
		case RemoteGameModelInterface.BLACK_HAS_WON:
			return "Black has won." ;
		case RemoteGameModelInterface.WHITE_HAS_WON:
			return "Black has won." ;
		case RemoteGameModelInterface.BLACK_TO_MOVE:
			return myColour == BLACK ? "Your move" : "Black to move" ;
		case RemoteGameModelInterface.WHITE_TO_MOVE:
			return myColour == WHITE ? "Your move" : "White to move" ;
		default: throw new IllegalStateException() ;
		}
	}
}
