package othello.view;

import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.*;

import javax.swing.JPanel;

import othello.model.GameConstants;
import othello.model.Pair;
import othello.model.RemoteGameModelInterface;

/**
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore Norvell
 * @version 1.0
 */

public class BoardView extends JPanel implements Observer, GameConstants {

    private static final long serialVersionUID = -8394619531996679312L;
	Animator animationModel ;
    ViewStrategy worldToViewMap ;
    Color gameStateColour = new Color(0x66, 0x66, 0x66, 128) ;
    Font gameStateFont = new Font(Font.SERIF, Font.BOLD, 20) ;
    
    
    public BoardView(final Animator animationModel, ViewStrategy vs ) {
        this.animationModel = animationModel ;
        this.worldToViewMap = vs ;
        animationModel.addObserver(this) ;
        addMouseListener(new MouseAdapter() {
        	@Override
        	public void mouseClicked(MouseEvent e) {
        		int x = e.getX(), y = e.getY() ;
        		int w = BoardView.this.getWidth(), h = BoardView.this.getHeight() ;
        		final Pair pair = worldToViewMap.viewToWorld(w, h, x, y) ;
        		animationModel.move( pair.getRow(), pair.getCol() ) ;
        	} }) ;
        }
        
    /** This implements the Observer interface
     */
    public void update(Observable o, Object arg) {
        repaint() ; }
    
    @Override protected void paintComponent( Graphics g ) {
    	super.paintComponent(g) ;
    	int width = getWidth(), height = getHeight() ;
    	Color origColour = g.getColor() ;
    	
		synchronized( animationModel ) {
	        Point2D p0, p1, p2, p3 ;
	        /* Paint the green background for the board. */ {
	            p0 = worldToView( width, height, 0.0, 0.0 ) ;
	            p1 = worldToView( width, height, 0.0, (double)COLS ) ;
	            p2 = worldToView( width, height, (double)ROWS, 0.0 ) ;
	            p3 = worldToView( width, height, (double)ROWS, (double)COLS ) ;
	            g.setColor(new Color(0, 200, 0) ) ;
	            paintFilledRectangle( g, p0, p1, p2, p3 ) ;}
	            
	        /* Draw vertical lines */ {
	            g.setColor(Color.black);
	            for( int c=0; c < COLS+1 ; ++c ) {
	                p0 = worldToView( width, height, 0.0, (float)c ) ;
	                p1 = worldToView( width, height, (float)COLS, (float)c ) ;
	                g.drawLine((int) p0.getX(), (int) p0.getY(), (int) p1.getX(), (int) p1.getY()) ; } }
	                
	        /* Draw horizontal lines */ {
	            g.setColor(Color.black);
	            for( int r=0; r < ROWS+1 ; ++r ) {
	                p0 = worldToView( width, height, (float)r, 0.0 ) ;
	                p1 = worldToView( width, height, (float)r, (float)COLS ) ;
	                g.drawLine((int)p0.getX(), (int)p0.getY(), (int)p1.getX(), (int)p1.getY()) ; } }
	        
	        /* Draw the static pieces */ {
	            for( int r=0 ; r < ROWS ; ++r ) {
	                for( int c=0 ; c < COLS ; ++c ) {
	                    int colour = animationModel.getPieceAt(r,c) ;
	                    if( colour != EMPTY ) {
	                        p0 = worldToView( width, height, r+0.1, c+0.1 ) ;
	                        p1 = worldToView( width, height, r+0.1, c+0.9 ) ;
	                        p2 = worldToView( width, height, r+0.9, c+0.9 ) ;
	                        p3 = worldToView( width, height, r+0.1, c+0.9 ) ;
	                        paintPiece( g, colour, p0, p1, p2, p3 ) ; } } } }
	        
	        /* Draw the title */ 
	            //paintString( g, -1.9, 0.0, -0.1, COLS, "OTHELLO" ) ;
	            //paintString( g, ROWS+1.9, COLS, ROWS+0.1, 0, "OTHELLO") ;
	        
	        /* Draw any flipping piece */ {
	            if( animationModel.getState() == animationModel.FLIPPING_STATE ) {
	                int colour = animationModel.getNewColour() ;
	                int r = animationModel.getFlippingRow() ;
	                int c = animationModel.getFlippingCol() ;
	                double amount = animationModel.getFlippingAmount() ;
	                if( amount < 0.5 ) {
	                    colour = colour==BLACK ? WHITE : BLACK ; }
	                else { 
	                    amount = 1 - amount ; }
	                // Now amount is between 0.5 and 1.
	                // When amount is 1, the bounding box is 0.8 by 0.8
	                // and centred in the rectangle.
	                p0 = worldToView( width, height, r+0.1+amount*0.8, c+0.1 ) ;
	                p1 = worldToView( width, height, r+0.1+amount*0.8, c+0.9 ) ;
	                p2 = worldToView( width, height, r+0.9-amount*0.8, c+0.9 ) ;
	                p3 = worldToView( width, height, r+0.9-amount*0.8, c+0.1 ) ;
	                paintPiece( g, colour, p0, p1, p2, p3 ) ; } }
	        
	        /* Draw any flying piece */ {
	            if( animationModel.getState() == animationModel.FLYING_STATE ) {
	                int colour = animationModel.getNewColour() ;
	                double r = animationModel.getFlyingRow() ;
	                double c = animationModel.getFlyingCol() ;
	                p0 = worldToView( width, height, r+0.1, c+0.1 ) ;
	                p1 = worldToView( width, height, r+0.1, c+0.9 ) ;
	                p2 = worldToView( width, height, r+0.9, c+0.9 ) ;
	                p3 = worldToView( width, height, r+0.1, c+0.9 ) ;
	                paintPiece( g, colour, p0, p1, p2, p3 ) ; } }
		
	       /* Draw  game information at the top of the screen */ {
	            String myColourString = animationModel.getColourString() ;
	            String moveString = animationModel.stateMessage() ;
	            
	            g.setColor(gameStateColour) ;
	            int x = width / 10 ;
	            int y = height / 10 ;
	            g.setFont(gameStateFont) ;
	            g.drawString(myColourString +" "+moveString, x, y) ; }
	            
	        g.setColor(origColour) ;
	    }
	}
                
            
            
    private void paintPiece( Graphics g, int colour,
                             Point2D p0, Point2D p1, Point2D p2, Point2D p3 ) {
        // Draw a piece more or less as defined by view space
    		// quadralateral (p0, p1, p2, p3)
            int left = (int) Math.min( p0.getX(), Math.min( p1.getX(), Math.min( p2.getX(), p3.getX() ) ) ) ;
            int right = (int) Math.max( p0.getX(), Math.max( p1.getX(), Math.max( p2.getX(), p3.getX() ) ) ) ;
            int top = (int) Math.min( p0.getY(), Math.min( p1.getY(), Math.min( p2.getY(), p3.getY() ) ) ) ;
            int bot = (int) Math.max( p0.getY(), Math.max( p1.getY(), Math.max( p2.getY(), p3.getY() ) ) ) ;
            int width = right-left ;
            int height = bot-top ;
            g.setColor( colour==BLACK ? Color.black  : Color.white ) ;
            g.fillOval(left, top, width, height);
            g.setColor( colour==BLACK ? Color.white  : Color.black );
            g.drawOval(left, top, width, height); }
    
    private void paintFilledRectangle( Graphics g,
                                       Point2D tl, Point2D tr, Point2D bl, Point2D br ) {
        // Draw a piece
            int [] xs = new int[]{ (int)tl.getX(), (int)tr.getX(), (int)br.getX(), (int)bl.getX() } ;
            int [] ys = new int[]{ (int)tl.getY(), (int)tr.getY(), (int)br.getY(), (int)bl.getY() } ;
            g.fillPolygon(xs, ys, 4); }
    
    private Point2D worldToView( double width, double height, double row, double col ) {
        return worldToViewMap.worldToView( width, height, row, col )  ; }
    
    /** Puts a string in the box defined by a rectangle in world space.
     * The strings origin is the bottom left corner of this box.
     * @param g
     * @param rTop
     * @param cLeft
     * @param rBot
     * @param cRight
     * @param message
     */
    private void paintString( Graphics g, double width, double height,
    		                              double rTop, double cLeft,
                                          double rBot, double cRight,
                                          String message  ) {
    	// Turn the world rectangle into a quadralateral in view space.
        Point2D p0 = worldToView( width, height, rTop, cLeft ) ;  // Top left
        Point2D p1 = worldToView( width, height, rTop, cRight ) ; // Top Right
        Point2D p2 = worldToView( width, height, rBot, cRight ) ; // Bottom Right
        Point2D p3 = worldToView( width, height, rBot, cLeft ) ;  // Bottom Left
        // Make points at the midpoints of segments p3--p2 and p0--p1.
        Point2D p2point5 = new Point2D.Double( (p2.getX()+p3.getX())/2, (p2.getY()+p3.getY())/2 ) ;
        Point2D p0point5 = new Point2D.Double( (p0.getX()+p1.getX())/2, (p0.getY()+p1.getY())/2 ) ;
        
        // It might look a bit like this
        //     p0---p.05---p1  
        //    /             |
        //   /              |  
        //  /               | 
        // p3-----p2.5-----p2  
        
        // Since fitting a string into an arbitrary quadralateral is
        // a bother, we instead fit it into a parallelagram defined
        // by line segment p3 to p2 (as base line) an angle alpha and
        // a height. Alpha is the angle from  p3--p2 to p2.5--p0.5.
        // If the string begins with letter T, its arm is parallel to p3-p2
        // and its stem is at an angle of alpha to the base-line. 
        //     -----------------  ---
        //    / Here alpha is /    |
        //   / about 70      /     |height
        //  / degrees       /      |
        // p3--------------p2     ---
        // The height is the length of p2.5--p0.5 times sin(alpha)
        // This won't look quite right in perspective transform, as
        // perspective demands a different shear and scale for each letter.
        // Better to give each letter its own transformation. 
        // However, this is good enough for government work.
        
        // Altogether there is a translation, a rotation, scalings, and a shear.
        
        // We write the string along the line from p3 to p2.
        // Theta is the angle this line makes with the x-axis
        double theta = Math.atan2(p2.getY()-p3.getY(), p2.getX()-p3.getX() ) ;
        // rho is the angle the line from p2.5 to p0.5 makes with the x axis.
        double rho = Math.atan2((p0point5.getY()-p2point5.getY()), (p0point5.getX()-p2point5.getX()) ) ;
        // Alpha is the angle from vector (p3,p2) to vector (p2.5,p0.5).
        // We'll use alpha to shear the font so it fits in the parallelegram.
        double alpha = theta - rho ;
        
        // TODO Here be dragons. On Macs there is a bug in Java
        // so that drawstring is incorrect on a rotated Graphics object.
        // See http://stackoverflow.com/questions/14767826/rotating-text-in-java2d-on-mac
        // for an approach that might work.
        
        // Need a new graphics context
        		Graphics2D g2d = (Graphics2D) g.create() ;
        // Translate p3 to the origin of the new context
        	    g2d.translate(p3.getX(), p3.getY()) ;
        // Rotate the new context so that p2 is on the x-axis.
        		g2d.rotate(theta) ;
        // Shear so that the letters slant the "right" amount
        		g2d.shear( -Math.cos(alpha) , 0) ;
        
        // Make a font
            Font font = new Font("SansSerif", Font.PLAIN, 100 ) ;
        // Calculate the height and width of the string in this font
            FontMetrics fm = g2d.getFontMetrics(font) ;
            int h = fm.getMaxAscent() ;
            int w = fm.stringWidth(message) ;
        // Calculate the scale factor to stretch the string from p3 to p2.
            double strWidth = Math.sqrt( (p3.getX()-p2.getX())*(p3.getX()-p2.getX())
            		                    +(p3.getY()-p2.getY())*(p3.getY()-p2.getY()) ) ;
            double xScale = strWidth / w ;
        // And so the letters are the right height.
            double strHeight = Math.sqrt( (p2point5.getX()-p0point5.getX())*(p2point5.getX()-p0point5.getX())
            		                  +(p2point5.getY()-p0point5.getY())*(p2point5.getY()-p0point5.getY()) ) ;
            strHeight = strHeight * Math.sin(alpha) ;
            double yScale = height / h ;
            
        g2d.scale( xScale, yScale ) ;
        
        
        g2d.setColor(Color.black);
        g2d.setFont( font );
        g2d.drawString(message, 0, 0 ) ;
        g2d.dispose() ; }
}